import os
import pandas as pd
from pandas.io.sql import SQLTable
import datetime
import sqlalchemy
import csv
from io import StringIO
from typing import Union, List
from . import Types
from . import Constants
from dotenv import load_dotenv
load_dotenv()


cnx = sqlalchemy.create_engine('postgresql://%s:%s@%s/%s' % (os.environ.get('POSTGRES-USER'),
                               os.environ.get('POSTGRES-PASS'), os.environ.get('POSTGRES-URL'), os.environ.get('POSTGRES-DB')))


def load_env_file(path: str):
    load_dotenv(path)
    globals()['cnx'] = sqlalchemy.create_engine('postgresql://%s:%s@%s/%s' % (os.environ.get('POSTGRES-USER'),
                                                                              os.environ.get('POSTGRES-PASS'), os.environ.get('POSTGRES-URL'), os.environ.get('POSTGRES-DB')))


def create_historical_table() -> None:
    cmd = '''
        CREATE TABLE symbols (
        symbol TEXT PRIMARY KEY,
        name TEXT
        );

        CREATE TABLE historical (
        time TIMESTAMP WITHOUT TIME ZONE NOT NULL,
        symbol TEXT,
        open NUMERIC,
        high NUMERIC,
        low NUMERIC,
        close NUMERIC,
        volume NUMERIC,
        FOREIGN KEY (symbol) REFERENCES symbols (symbol),
        UNIQUE (time, symbol)
        );

        SELECT create_hypertable('historical', 'time');

        create index on historical (symbol, time desc);
        '''
    cnx.execute(cmd)


def write_symbols(symbols: List[Types.Symbol]) -> None:
    formated_symbols = ',\n'.join("('%s')" % i.upper() for i in symbols)
    cmd = 'INSERT INTO symbols (symbol) VALUES\n%s\n' % formated_symbols
    cmd += 'ON CONFLICT DO NOTHING;'
    cnx.execute(cmd)


def get_symbols() -> List[Types.Symbol]:
    query = 'SELECT * FROM symbols;'
    symbols = pd.read_sql(query, con=cnx).symbol.to_list()
    return sorted(symbols)


def write_historical(sf: Types.StockFrame) -> None:
    def historical_method(table: SQLTable, conn: sqlalchemy.engine.Engine, keys: list, dataIter):
        with conn.connection.cursor() as cur:
            sBuf = StringIO()
            writer = csv.writer(sBuf)
            writer.writerows(dataIter)
            sBuf.seek(0)

            columns = ', '.join('"%s"' % k for k in keys)
            if table.schema:
                table_name = '%s.%s' % (table.schema, table.name)
            else:
                table_name = table.name
            temp_table_name = 'temp_%s' % table_name
            update_set = ", ".join(
                ['%s=EXCLUDED.%s' % (v, v) for v in keys if v not in ['time', 'symbol']])

            cmd = 'CREATE TEMPORARY TABLE %s (LIKE %s) ON COMMIT DROP;' % (
                temp_table_name, table_name)
            cur.execute(cmd)

            cmd = 'COPY %s (%s) FROM STDIN WITH CSV' % (
                temp_table_name, columns)
            cur.copy_expert(sql=cmd, file=sBuf)

            cmd = 'INSERT INTO %s(%s)\nSELECT %s FROM %s\n' % (
                table_name, columns, columns, temp_table_name)
            cmd += 'ON CONFLICT (time, symbol) DO UPDATE SET %s;' % update_set
            cur.execute(cmd)

            cur.execute('DROP TABLE %s' % temp_table_name)

    sf.index.names = ['time']
    start_timer = datetime.datetime.now()
    write_symbols(sf['symbol'].unique())
    sf.to_sql('historical', con=cnx, if_exists='append',
              index=True, method=historical_method)
    print(sf.symbol[0], len(sf.index), datetime.datetime.now() - start_timer)
    sf.index.names = ['timestamp']


def get_historical(symbols: Union[List[Types.Symbol], None] = None, start: Union[Types.Datetime, None] = None, end: Union[Types.Datetime, None] = None, time_list: Union[List[datetime.datetime], None] = None) -> Types.StockFrame:
    def get_query_date(date: datetime.datetime, end: bool = False):
        if end:
            return date.strftime('%Y-%m-%d 23:59:00')
        else:
            return date.strftime('%Y-%m-%d 00:00:00')

    binder = 'WHERE'
    query_filter = ''
    if symbols is not None and len(symbols) > 0:
        query_filter += 'WHERE symbol in (%s)\n' % (
            ','.join("'%s'" % i.upper() for i in symbols))
        binder = 'AND'
    if start is not None:
        if end is not None:
            query_filter += "%s time BETWEEN '%s' AND '%s'\n" % (
                binder, get_query_date(start), get_query_date(end, True))
        else:
            query_filter += "%s time >= '%s'\n" % (binder,
                                                   get_query_date(start))
        binder = 'AND'
    elif end is not None:
        query_filter += "%s time <= '%s'\n" % (
            binder, get_query_date(end, True))
        binder = 'AND'
    if time_list is not None:
        query_filter += '%s time::time in (%s)' % (binder,
                                                   ', '.join(i.strftime("'%H:%M:00'") for i in time_list))

    query = 'SELECT %s FROM historical\n%s' % (
        ','.join('%s' % i for i in Constants.stonkDb_colums), query_filter)

    start_timer = datetime.datetime.now()
    s_buf = StringIO()
    conn = cnx.raw_connection()
    cur = conn.cursor()
    cmd = "COPY (%s) TO STDOUT WITH (FORMAT CSV, DELIMITER ',')" % query
    cur.copy_expert(cmd, s_buf)
    s_buf.seek(0)

    sf = pd.read_csv(s_buf, names=Constants.stonkDb_colums, parse_dates=[
                     'time']).set_index('time').sort_index()
    sf.index.names = ['timestamp']
    # print(len(sf.index), datetime.datetime.now() - start_timer)
    return sf


def get_historical_date_range():
    start = cnx.execute('SELECT MIN(time) FROM historical').scalar()
    end = cnx.execute('SELECT MAX(time) FROM historical').scalar()
    return start, end


def get_historical_date_list():
    date_list = sorted([i[0] for i in cnx.execute(
        'SELECT DISTINCT time FROM historical')])
    return date_list



import datetime
import os
import requests
import pandas as pd
import time
import random
from tqdm import tqdm
from multiprocessing import Pool
from dotenv import load_dotenv
from typing import Union, List
from . import Types
load_dotenv()


tradier_url = 'https://api.tradier.com'


def load_env_file(path: str):
    load_dotenv(path)


def _get_historical(symbol: Types.Symbol, start: Union[None, Types.Datetime] = None, end: Union[None, Types.Datetime] = None, rate_limit: bool = False):
    def process_frame(sf: Types.StockFrame, symbol: Types.Symbol):
        sf = sf.dropna()
        sf = sf.rename(columns={'date': 'timestamp'})
        sf['symbol'] = symbol.upper()
        sf['timestamp'] = pd.to_datetime(sf['timestamp'], utc=True)
        sf['timestamp'] = sf['timestamp'].dt.tz_convert('US/Eastern')
        sf['timestamp'] = sf['timestamp'].apply(lambda x: (
            x + datetime.timedelta(hours=12)).replace(hour=16, minute=30))
        sf = sf.drop_duplicates(subset=['timestamp'])
        sf = sf.set_index('timestamp')
        return sf

    if rate_limit:
        time.sleep(random.randrange(1, 20) / 10.0)

    url = '%s/v1/markets/history' % tradier_url
    params = {
        'symbol': symbol.upper(),
        'interval': 'daily',
    }
    if start is not None:
        params['start'] = start.strftime('%Y-%m-%d')
    if end is not None:
        params['end'] = end.strftime('%Y-%m-%d')

    headers = {
        'Authorization': 'Bearer %s' % os.getenv('TRADIER_TOKEN'),
        'Accept': 'application/json'
    }

    response = requests.get(url, params=params, headers=headers)
    try:
        json_response = response.json()
        sf: Types.StockFrame = pd.DataFrame(json_response['history']['day'])
        sf = process_frame(sf, symbol)
        return sf
    except:
        print('%s from %s to %s failed' %
              (symbol, params['start'], params['end']))
        return None


def get_historical(symbols: List[Types.Symbol], start: Union[None, Types.Datetime] = None, end: Union[None, Types.Datetime] = None):
    jobs = []
    pool = Pool(processes=10)

    for symbol in symbols:
        jobs.append(pool.apply_async(func=_get_historical,
                    args=(symbol, start, end, True)))
    pool.close()
    results = []
    for job in tqdm(jobs):
        results.append(job.get())

    return [i for i in results if i is not None]



from . import StonkDataLib as SDL
from . import utils
from . import Types
from . import Constants
import pandas as pd
import random
from typing import Union
import numpy as np
import datetime


def set_seed(seed: int):
    random.seed(seed)
    np.random.seed(seed)


def load_env_file(path: str):
    SDL.load_env_file(path)


class Env:
    def __init__(self, start_balance: float, min_date: Union[datetime.datetime, None] = None, max_date: Union[datetime.datetime, None] = None) -> None:
        self.lookback_period = 90
        self.episode_period = 90
        # self.lookback_period = 25
        # self.episode_period = 22
        self.trade_period = 3
        self.transaction_fee = 0
        self.transaction_threshold = 1e-4
        self.done_treashold = 100
        self.start_balance = start_balance
        self.min_date, self.max_date = min_date, max_date

        self.reset()

    def reset(self):
        self.balance = self.start_balance * 1
        self.last_balance = self.balance * 1
        self.set_start_date()
        self.set_stock_frame()
        state = self.get_state()
        return state

    def set_start_date(self) -> None:
        date_list = SDL.get_historical_date_list()
        if self.min_date is not None:
            date_list = [i for i in date_list if i.date() >=
                         self.min_date.date()]

        if self.max_date is not None:
            date_list = [i for i in date_list if i.date() <=
                         self.max_date.date()]

        idx = random.choice(list(range(len(date_list)))[
                            self.lookback_period:-self.episode_period])

        self.date = date_list[idx]
        self.start_date = date_list[idx - self.lookback_period]
        self.end_date = date_list[idx + self.episode_period]

        self.date_list = date_list[(
            idx - self.lookback_period):(idx + self.episode_period + 1)]
        self.index = self.lookback_period

    def process_stock_frame(self, sf: Types.StockFrame) -> Types.StockFrame:

        def group_processing(subframe):
            close = subframe['close']
            low = subframe['low']
            high = subframe['high']
            subframe = subframe.sort_index()
            subframe['rsi'] = utils.rsi(close, 14)
            subframe['macd'] = utils.macd(close)
            subframe['cci'] = utils.cci(close, high, low, 20)

            return subframe

        sf = sf.groupby('symbol').apply(group_processing)
        sf = sf.sort_values(by=['timestamp', 'symbol'])
        sf = sf.fillna(0)

        return sf

    def initialize_shares(self) -> None:
        initial_weight = 1.0 / len(self.symbols)

        def set_shares(subframe: Types.StockFrame):
            subframe = subframe.sort_index()
            subframe['shares'] = initial_weight / subframe['close'][0]
            subframe['value'] = subframe['shares'] * subframe['close']
            return subframe

        sf = self.sf
        sf = sf.groupby('symbol').apply(set_shares)
        sf = sf.sort_values(by=['timestamp', 'symbol'])

        sf.loc[sf.index > self.date, ['shares']] = 0
        self.sf = sf.drop(columns=['value'])

    def set_stock_frame(self) -> None:
        self.sf = SDL.get_historical(
            start=self.start_date, end=self.end_date)

        self.sf = self.sf.sort_values(by=['timestamp', 'symbol'])
        self.sf['day'] = self.sf.index.floor('d')
        self.symbols = self.sf['symbol'].unique()

        self.initialize_shares()
        self.sf = self.process_stock_frame(self.sf)
        assert len(self.sf) == len(self.symbols) * len(self.date_list)

    def get_viewable_stock_frame(self) -> Types.StockFrame:
        dates = self.date_list[(
            self.index - self.lookback_period): (self.index + 1)]
        sf = self.sf[self.sf.index.isin(dates)]
        return sf

    def get_state(self):
        '''
        returns a (m, n, b) sized array
        m dimension: symbol
        n dimension: lookback period
        b dimension: symbol attributs ['weight', 'open', 'high', 'low', 'close', 'volume', 'rsi', 'macd', 'cci', 'adx']
        '''
        sf = self.get_viewable_stock_frame()
        sf = sf.sort_values(['symbol', 'day'])

        group = sf.reset_index(drop=True).groupby(['symbol'])
        data_columns = ['shares'] + Constants.sf_data_columns + \
            ['rsi', 'macd', 'cci']
        raw_data = sf[data_columns].values
        data = np.array([raw_data[i.values, :]
                        for k, i in group.groups.items()])

        # normalize data
        data_min = np.repeat(
            data.min(axis=1)[:, :, np.newaxis], data.shape[1], axis=2).swapaxes(2, 1)
        data_max = np.repeat(
            data.max(axis=1)[:, :, np.newaxis], data.shape[1], axis=2).swapaxes(2, 1)
        scaled = (data - data_min) / (data_max - data_min + 1e-16)

        return scaled

    def step(self, action: np.ndarray):
        def backfill_shares(subframe):
            subframe = subframe.sort_index()
            subframe_slice = subframe[subframe.index <= self.date]
            idx = subframe_slice['shares'] == 0
            last_shares = subframe_slice.loc[idx == False, 'shares'][-1]
            subframe_slice.loc[idx, 'shares'] = last_shares
            subframe[subframe.index <= self.date] = subframe_slice
            return subframe

        # get old values
        old_price = self.sf.loc[self.sf.index == self.date, 'close'].values
        old_shares = self.sf.loc[self.sf.index == self.date, 'shares'].values

        # advance date
        done = False
        if (self.index + self.trade_period * 2) >= len(self.date_list):
            self.index = len(self.date_list) - 1
            done = True
        else:
            self.index += self.trade_period
        self.date = self.date_list[self.index]

        # calculate new shares from given weights (action)
        new_price = self.sf.loc[self.sf.index == self.date, 'close'].values
        new_shares = action / new_price
        self.sf.loc[self.sf.index == self.date, 'shares'] = new_shares

        # calculate delta-shares for transaction free
        idx = abs(old_shares - new_shares) <= self.transaction_threshold
        transaction_fee = sum(idx == False) * self.transaction_fee

        # calculate reward using new balance
        new_balance = self.balance * \
            (1 + (old_shares * (new_price - old_price)).sum()) - transaction_fee
        reward = new_balance - self.balance
        self.balance = new_balance

        # backfill shares if necessary
        if self.trade_period > 1:
            sf = self.sf
            sf = sf.groupby('symbol').apply(backfill_shares)
            sf = sf.sort_values(by=['timestamp', 'symbol'])
            self.sf = sf

        # get new state
        state = self.get_state()

        # check if done
        if self.balance <= self.done_treashold:
            done = True
            reward = -1e4

        return state, reward, done



import datetime
import pandas as pd

epsilon = 1e-16


def rsi(series: pd.Series, n: int = 14) -> pd.Series:
    delta = series.diff()
    up = delta.clip(lower=0)
    down = -delta.clip(upper=0)
    ema_up = pd.Series.ewm(up, span=n).mean()
    ema_down = pd.Series.ewm(down, span=n).mean() + epsilon
    rsi = ema_up / ema_down
    rsi = 100 - (100 / (1 + rsi))

    return rsi + epsilon


def macd(series: pd.Series) -> pd.Series:
    exp1 = series.ewm(span=12, adjust=False).mean()
    exp2 = series.ewm(span=26, adjust=False).mean()
    macd = exp1 - exp2
    return macd + epsilon


def cci(close: pd.Series, high: pd.Series, low: pd.Series, n: int = 20) -> pd.Series:
    typical_price = (high + low + close) / 3.0
    moving_average = typical_price.rolling(n).mean()
    mean_deviation = typical_price.rolling(n).std()
    cci = (typical_price - moving_average) / (0.15 * mean_deviation + epsilon)
    return cci + epsilon


# def get_adx(close: pd.Series, high: pd.Series, low: pd.Series, n: int = 14) -> pd.Series:
#     plus_dm = high.diff().clip(lower=0)
#     minus_dm = low.diff().clip(upper=0)

#     tr1 = pd.DataFrame(high - low)
#     tr2 = pd.DataFrame(abs(high - close.shift(1)))
#     tr3 = pd.DataFrame(abs(low - close.shift(1)))
#     frames = [tr1, tr2, tr3]
#     tr = pd.concat(frames, axis=1, join='inner').max(axis=1)
#     atr = tr.rolling(n).mean()

#     plus_di = 100 * (plus_dm.ewm(alpha=1 / n).mean() / atr)
#     minus_di = abs(100 * (minus_dm.ewm(alpha=1 / n).mean() / atr))
#     dx = (abs(plus_di - minus_di) / abs(plus_di + minus_di)) * 100
#     adx = ((dx.shift(1) * (n - 1)) + dx) / n
#     adx_smooth = adx.ewm(alpha=1 / n).mean()
#     return plus_di, minus_di, adx_smooth


def get_sp500_filtered_symbols(date: datetime.datetime = datetime.datetime(year=2008, month=1, day=1), n: int = 50):
    def clean_date(x):
        if type(x) is str:
            x = x.split(' ')[0].strip()
        else:
            x = '2021-01-01'
        return x
    payload = pd.read_html(
        'https://en.wikipedia.org/wiki/List_of_S%26P_500_companies')
    sp500 = payload[0]
    sp500 = sp500.rename(columns={'Date first added': 'date_added'})
    sp500 = sp500[['Symbol', 'date_added', 'Founded']]

    payload = pd.read_html(
        'file:///home/hotdog/Downloads/sp500_market_cap.html')
    market_cap = payload[0].sort_values('Weight', ascending=False)
    market_cap = market_cap[['Symbol', 'Weight']]

    sp500 = sp500.merge(market_cap)
    sp500['date_added'] = sp500['date_added'].apply(clean_date)
    sp500['date_added'] = pd.to_datetime(sp500['date_added'])

    filtered = sp500[sp500['date_added'] <=
                     date].sort_values('Weight', ascending=False)

    return filtered.iloc[:n]['Symbol'].tolist()



from setuptools import find_packages, setup

setup(
    name='MarketGym',
    version='0.1.2',
    description='Python Reinforcement Environment for Stock Trading',
    url='https://github.com/jasonpul/MarketGym',
    author='Jason Pul',
    license='MIT',
    packages=['MarketGym'],
    include_package_data=True,
    install_requires=['certifi==2021.5.30', 'charset-normalizer==2.0.4', 'greenlet==1.1.0', 'idna==3.2', 'lxml==4.6.3', 'numpy==1.21.1', 'pandas==1.3.1', 'plotly==5.1.0', 'psycopg2==2.9.1',
                      'python-dateutil==2.8.2', 'python-dotenv==0.19.0', 'pytz==2021.1', 'requests==2.26.0', 'six==1.16.0', 'SQLAlchemy==1.4.22', 'tenacity==8.0.1', 'tqdm==4.62.0', 'urllib3==1.26.6'],
    keywords=['reinforcement', 'stock', 'postgres', 'tradier'],
)
